/*   utils.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-2000 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

static const char _rcsid[] = "$Id: utils.c,v 1.11.2.2 2003/04/09 07:38:33 elmoe Exp $";

#include "commondef.h"

#include "utils.h"

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef DES_ENCRYPTION
#include "des.h"
#endif

#ifdef TLS
#include "../ftps/tlsutil.h"

#define FGETS tls_fgets
#else
#define FGETS fgets
#endif

gint strip_crlf (gchar buffer[])
{
  gint n = 0;
  gchar *ptr;

  g_return_val_if_fail (buffer != NULL, -1);

  ptr = buffer + strlen(buffer) - 1;
  
  while ((ptr > buffer) && ((*ptr == '\n') || (*ptr == '\r')))
    {
      ptr--;
      n++;
    }
  
  *(ptr+1) = '\0';
  
  return n;
}

gchar skipspace(gchar **str)
{
  g_return_val_if_fail (str != NULL && *str != NULL, '\0');
 
  while (**str && isspace (**str)) 
    (*str)++;

  return **str;
}

static gboolean
isquote (gchar c)
{
  return (c == '\'' || c == '"' || c == '`' || c == '');
}

gchar skipnspace(gchar **str)
{ 
  gboolean marked = FALSE;

  g_return_val_if_fail (str != NULL && *str != NULL, '\0');
   
  while (**str) {
    if (isquote (**str))
      marked = !marked; /* fixme -- should match the opening quote */
    
    if(!isspace(**str) || marked) 
      (*str)++;
    else
      break;
   }

   return **str;
}

gchar skiptocolon (gchar **str)
{
  g_return_val_if_fail (str != NULL && *str != NULL, '\0');

  skipspace(str); /* skip leading whitespaces */
  
  while (**str && !isspace (**str) && (**str != ':'))
    (*str)++; /* skip while neither space nor ':' */

  return **str;
}

gchar skipspctocolon (gchar **str)
{ 
  g_return_val_if_fail (str != NULL && *str != NULL, '\0');

  skipspace(str); /* skip whitespaces */
  
  if (**str != ':')
    return **str;

  (*str)++; /* move over the colon */
  skipspace(str); /* skip leading whitespaces of next field */
  
  return **str;
}

int readline(char* line, int linelen, FILE* file)
{ char* r;

  r = fgets(line,linelen,file);
  if (r) strip_crlf(line);
  return (r != NULL);
}

/* creates filename relative to server_dir */
void
mk_filename(gchar *buf, const gchar filename[], const gchar defaultname[])
{
  extern gchar server_dir[];

  g_return_if_fail (buf != NULL);

  if (!filename)
    filename = defaultname;
  
  if (!filename)
    {
      strcpy (buf, server_dir);
      return;
    }

  if (g_path_is_absolute (filename))
    strcpy (buf, filename);
  else
    {
      strcpy(buf,server_dir);
      strcat(buf,filename);
    }
}

/* read with timeout */
int read_to(int fd, char* buf, int buflen, int timeout)
{ fd_set         fdset;
  int            i;
  struct timeval to;

  to.tv_sec = timeout;
  to.tv_usec = 0;
  FD_ZERO(&fdset);
  FD_SET(fd,&fdset);
  i = select(fd+1,&fdset,NULL,NULL,timeout ? &to : NULL);
  if (i > 0) i = (int)read(fd,buf,buflen); else
  { if (!i)
    { errno = ETIMEDOUT;
      i = -1;
    }
    *buf = '\0';
  }
  return i;
}

/* fgets with timeout */
char* fgets_to(char* buf, int buflen, FILE* stream, int timeout)
{ int            fd;
  fd_set         fdset;
  char*          ptr;
  struct timeval to;

  fd = fileno(stream);
  to.tv_sec = timeout;
  to.tv_usec = 0;
  FD_ZERO(&fdset);
  FD_SET(fd,&fdset);
  fd = select(fd+1,&fdset,NULL,NULL,timeout ? &to : NULL);
  if (fd > 0) ptr = FGETS(buf,buflen,stream); else
  { if (!fd) errno = ETIMEDOUT;
    *buf = '\0';
    ptr = NULL;
  }
  return ptr;
}

/* set flag */
void 
flag_set (guint32 *flags, guint n, gboolean value)
{ 
  g_return_if_fail (flags != NULL);
  g_return_if_fail (n < (sizeof (guint) << 3));

  if (value)
    *flags |= (1 << n);
  else
    *flags &= ~(1 << n);
}

/* get flag */
gboolean
flag_get (guint32 flags, guint n)
{ 
  g_return_val_if_fail (n < (sizeof (guint) << 3), FALSE);
  g_assert (TRUE == 1); /* paranoia */

  return (flags & (1 << n)) ? TRUE : FALSE;
}

/* accept with timeout */
int accept_to(int s, struct sockaddr *addr, int *addrlen, int timeout)
{ fd_set         fdset;
  int            n;
  struct timeval to;

  to.tv_sec = timeout;
  to.tv_usec = 0;
  FD_ZERO(&fdset);
  FD_SET(s,&fdset);
  n = select(s+1,&fdset,NULL,NULL,timeout ? &to : NULL);
  if (!n) { 
    errno = ETIMEDOUT;
    n  = -1;
  } else if (n > 0) 
    n = accept(s,addr,addrlen);

  return n;
}

/* connect with timeout */
int connect_to(int s, struct sockaddr *addr, int addrlen, int timeout)
{ fd_set         fdset;
  int            flags;
  int            n;
  struct timeval to;

  if ((flags = fcntl(s,F_GETFL)) >= 0) fcntl(s,F_SETFL,flags | O_NONBLOCK);
  n = connect(s,addr,addrlen);
  if ((n < 0) && (errno == EINPROGRESS))
  { to.tv_sec = timeout;
    to.tv_usec = 0;
    FD_ZERO(&fdset);
    FD_SET(s,&fdset);
    n = select(s+1,NULL,&fdset,NULL,timeout ? &to : NULL);
    if (!n)
    { errno = ETIMEDOUT;
      n = -1;
    } else if (n > 0) n = 0;
  }
  if (flags >= 0) fcntl(s,F_SETFL,flags);
  return n;
}

void sprint_difftime(char *str, int len, double diff)
{
   int    d = 0, 
          h = 0, 
	  m = 0;
 
   d = diff / 86400;
   h = (diff - d * 86400) / 3600;
   m = ((diff - d * 86400 - h * 3600) / 60);

   if (d > 0) {
      g_snprintf(str, len, "%d day%s, %d hour%s, %d minute%s", 
	 d, (d == 1 ? "" : "s"), 
	 h, (h == 1 ? "" : "s"),
	 m, (m == 1 ? "" : "s"));
   }
   else if (h > 0) {
      g_snprintf(str, len, "%d hour%s, %d minute%s", 
	 h, (h == 1 ? "" : "s"),
	 m, (m == 1 ? "" : "s"));  
   }
   else {
      g_snprintf(str, len, "%d minute%s", m, (m == 1 ? "" : "s"));   
   }
}

/* EOF */
