/* ---------------------------------------------------------------------------
 *
 *   This file is part of OpenFTPD (based on FTP4ALL)
 *   (C) 1996-2000 by Crescent, (C) 2000 by the OpenFTPD Team
 *
 * ---------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 * ---------------------------------------------------------------------------
 *   MODULE: dupecheck.c - check for dupes
 * ---------------------------------------------------------------------------
 */

#include "ac_config.h"
 
#include <sys/types.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <regex.h>
#include <fnmatch.h>
#include <time.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include "externsd.h"
#include "log.h"
#include "serverd.h"
#include "utils.h"    
#include "user.h"
#include "group.h"
#include "strlcpy.h" 

enum {
   SEARCH_NORMAL,
   SEARCH_GLOB,   
   SEARCH_REGEX,
   SEARCH_EXACT
};


struct sf_ident_s
{ 
   char id[32];
   int  version;
   int  count;
   int  temp1;
   int  temp2;
} d_id = { "OpenFTPD dupe file", 1, 0, 0, 0 };


int alloc_dupe(void)
{ 
   void *ptr = NULL;

   ptr = g_realloc(dupe, (dupe_size + 1) * sizeof(struct dupe_s));
   if (!ptr) return dupe_size;
   dupe = ptr;
   dupe_size++;
   return dupe_size;
}


/*****************************************************************************
 * open dupelog, if new = 0 check header for version, else create a new file *
 * and write the header                                                      *
 *****************************************************************************/
FILE *open_dupefile(char *mode, int new)
{
   char                str[OPENFTPD_PATH_MAX];
   FILE                *file;
   struct sf_ident_s   id_test;

   mk_filename(str, DFL_DUPE_LOG, NULL);
   file = fopen(str, mode);
   if (!file) { 
      errlog("fopen(%s): %s (%i)", str, strerror(errno), errno);
      return NULL;
   }
   else {
      rewind(file);
      if (new) {
	 if (fwrite(&d_id, sizeof(struct sf_ident_s), 1, file) != 1) {
            errlog("fwrite(dupe header): %s (%i)", strerror(errno), errno);
            return NULL;
	 }    
      }
      else {
	 if (!fread(&id_test, sizeof(id_test), 1, file)) { 
	    fclose(file);
	    errlog("fread(dupe file): %s (%i)",strerror(errno), errno);
	    return NULL;
	 }
	 if (strcmp(id_test.id, d_id.id)) { 
	    errlog("dupe file %s has bad ID", str);
	    fclose(file);
	    return NULL;
	 } 
	 if (id_test.version > d_id.version) { 
	    errlog("dupe file is newer version than daemon");
	    fclose(file);
	    return NULL;
	 }
	 if (id_test.version < d_id.version) { 
	    errlog("dupe file is older version than daemon");
	    fclose(file);
	    return NULL;
	 }
      }
   }
   return file;
}



int is_dupe_detected(int type, char *pattern, char *fullpath, regex_t *preg)
{
   char *filename;
   
   filename = get_filename(fullpath);
   switch(type) {
      case SEARCH_NORMAL:    
	 if (filename && pattern) {
  	    if (strstr(filename, pattern)) return 1;
         }
	 break;
      case SEARCH_GLOB:    
	 if (filename && pattern) {
  	    if (!fnmatch(pattern, filename, 0)) return 1;
         }
	 break;
      case SEARCH_REGEX:    
	 if (fullpath && preg) {
	    if (!regexec(preg, fullpath, 0, NULL, 0)) return 1;
         }
	 break;
      case SEARCH_EXACT:
         if (pattern && fullpath) {
	    if (!strcmp(pattern, fullpath)) return 1;
         }
	 break;
   }
   return 0;
}



int f4adp_dupe_add(struct child_s *c, char **p, int n)
{
   struct dupe_s  *d;
   unsigned long  crc;
   time_t         t;

   alloc_dupe();
   if (dupe_count >= dupe_size) { 
      errlog("out of memory (dupes 0)");
      server_reply(c, "-out of memory");
      return 0;
   }
   d = dupe + dupe_count;
   dupe_count++;
   memset(d, 0, sizeof(struct dupe_s));
   strlcpy(d->filename, *(p+1), sizeof(d->filename));
   if (sscanf(*(p), "%lu", &crc)) d->crc32 = crc;
   time(&t);
   d->ul_time = t;
   d->ul_uid = c->user->uid;

   dupe_dirty = 1;
   server_reply(c, "+ok");
   return 0;
}


/*******************************************************************
 * search for matching entry in the dupelog and mark it as unduped *
 * the search is done either by strcasecmp() if no parameter was   *
 * given, by shell expanding (glob) if first argument is -g or by  *
 * extended regular expressions if first argument is -e            *
 *******************************************************************/
int f4adp_dupe_mark_unduped(struct child_s *c, char **p, int n)
{
   int                 i;
   long                files = 0;
   FILE                *file;
   char                str[1024];
   struct dupe_s       dp;
   regex_t             preg;
   int                 type = SEARCH_NORMAL;
   char                *pattern;
   int                 found;

   pattern = p[0];
   if (n == 2 && !strcasecmp(p[0], "-e")) {
      if (regcomp(&preg, p[1], REG_EXTENDED|REG_NEWLINE|REG_ICASE) != 0) {
         server_reply(c, "-Invalid regular expression.");
	 return 0;
      }
      type = SEARCH_REGEX;
      pattern = NULL;
   }
   else if (n == 2 && !strcasecmp(p[0], "-g")) {
      type = SEARCH_GLOB;
      pattern = p[1];
   }
   else if (n == 2 && !strcasecmp(p[0], "-x")) {
      type = SEARCH_EXACT;
      pattern = p[1];
   } 

   /* first undupe from memory ... */
   for (i = 0; i < dupe_count; i++) {
      found = is_dupe_detected(type, pattern, (dupe+i)->filename, &preg);
      if (found && ((dupe+i)->unduped == 0)) {
         (dupe+i)->unduped = 1;
         files++;
      }
   }
   
   /* ... then undupe in dupefile */
   file = open_dupefile("r+", 0);
   if (file) {
      while(!feof(file)) {
	 if (fread(&dp, sizeof(dp), 1, file)) {
            found = is_dupe_detected(type, pattern, dp.filename, &preg);
            if (found && (dp.unduped == 0)) {
               dp.unduped = 1;
               if (!fseek(file, - sizeof(struct dupe_s), SEEK_CUR)) {
        	  fwrite(&dp, sizeof(struct dupe_s), 1, file);
	       }
	       files++;
	    }
	 }
      }      
      fclose(file);
   }
   g_snprintf(str, sizeof(str), "+%ld", files);
   server_reply(c, str);
   return 0;
}


/*******************************************************************
 * search for matching entry in the dupelog and mark it as deleted *
 * the search is done by filename and path                         *
 *******************************************************************/
int f4adp_dupe_mark_deleted(struct child_s *c, char **p, int n)
{
   int                 i;
   long                files = 0;
   FILE                *file;
   char                str[128];
   struct dupe_s       dp;
 
   for (i = 0; i < dupe_count; i++) {
      if (!strcasecmp((dupe+i)->filename, *p) && (dupe+i)->deleted == 0) {
         (dupe+i)->deleted = 1;
         files++;
      }
   } 
   file = open_dupefile("r+", 0);
   if (file) {
      while(!feof(file)) {
	 if (fread(&dp, sizeof(dp), 1, file)) {
	    if (!strcasecmp(dp.filename, *p) && dp.deleted == 0) {
               dp.deleted = 1;
               if (!fseek(file, - sizeof(struct dupe_s), SEEK_CUR)) {
        	  fwrite(&dp, sizeof(struct dupe_s), 1, file);
	       }
	       files++;
	    }
         }
      }      
   }
   g_snprintf(str, sizeof(str), "+%ld", files);
   if(file) fclose(file);
   server_reply(c, str);
   return 0;
}


/********************************************************************
 * dump the dupes which are still kept in memory to the dupelog     *
 ********************************************************************/
int save_dupe_file()
{
   FILE*               file;
   int                 i;

   file = open_dupefile("r+", 0);
   if (file) {  /* dupelog was found and is valid */
      fseek(file, 0, SEEK_END);
   }
   else { /* dupefile does not exists yet or is invalid */
      file = open_dupefile("a+", 1);
      if (!file) return -1;     
   }
   for (i = 0; i < dupe_count; i++) {
      if (((dupe+i)->unduped == 0) || ((dupe+i)->deleted == 0)) {
	 if (!fwrite(dupe+i, sizeof(struct dupe_s), 1, file)) {
	    errlog("fwrite(dupe file): %s (%i)", strerror(errno), errno);
	    return -1; 
	 }
      }
   }
   fclose(file);

   /* free the dupelist in memory */
   g_free(dupe);
   dupe = NULL;
   dupe_count = 0;
   dupe_size = 0;
   dupe_dirty = 0;
   return 0;
}



/*************************************************************************
 * search for dupes in the binary dupelog and in memory by filename      *
 * and checksum of the first block of the file                           *
 *************************************************************************/
int f4adp_dupe_check(struct child_s *c, char **p, int n)
{ 
   FILE*               file;
   char                str[512];
   struct dupe_s       dp;
   unsigned long       crc32;
   struct user_s       *u = NULL;
   struct group_s      *g = NULL;
   struct child_s      *ch = NULL;
   time_t              t;
   int                 i;
   char                *fn;
   char                *fn2;

   sscanf(p[0], "%lu", &crc32);

   /* first we check the files currently uploaded (only by filename) */
   fn = get_filename(*(p+1));
   ch = child;
   for (i = 0; i < children; i++, ch++) {
      if (ch->user) g = get_group_by_gid(ch->user->gid);
      if (!strcasecmp(ch->filename, fn)) {
	 snprintf(str, sizeof(str), "-%s\t%s\t%ld\t\"%s\"",
	    ch->user->name, g ? g->name : "unknown", (long) -1, ch->cwd);
	 server_reply(c, str);
	 return 0;     
      }
   }
   
   /* then we check the dupes that are still in memory ... */
   for (i = 0; i < dupe_count; i++) {
      fn2 = get_filename((dupe+i)->filename);
      if (!strcasecmp(fn, fn2) && ((dupe+i)->crc32 == crc32) && 
         (dupe+i)->unduped == 0) {
         u = get_user_by_uid((dupe+i)->ul_uid);
	 if (u) g = get_group_by_gid(u->gid);
         time(&t);
	 g_snprintf(str, sizeof(str), "-%s\t%s\t%ld\t\"%s\"",
	    u ? u->name : "unknown", g ? g->name : "unknown", 
	    (long) (t - (dupe+i)->ul_time), (dupe+i)->filename);
	 server_reply(c, str);
	 return 0;
      }   
   }

   /* ... finally we check the dupelog file, if the dupefile is not
      ok we'll dump a warning to the errorlog and allow the file */
   file = open_dupefile("rb", 0);
   if (file) {
      while(!feof(file)) {
	 if (fread(&dp, sizeof(dp), 1, file)) {
            fn2 = get_filename(dp.filename);
	    if (!strcasecmp(fn, fn2) && (dp.crc32 == crc32) && 
	       dp.unduped == 0) {
               u = get_user_by_uid(dp.ul_uid);
	       if (u) g = get_group_by_gid(u->gid);
               time(&t);
	       snprintf(str, sizeof(str), "-%s\t%s\t%ld\t\"%s\"",
		  u ? u->name : "unknown", g ? g->name : "unknown", 
		  (long) (t - dp.ul_time), dp.filename);
	       server_reply(c, str);
               fclose(file);
	       return 0;
	    }
         }
      }      
      fclose(file);
   }
   server_reply(c, "+ok");
   return 0;
}



int f4adp_dupe_find_file(struct child_s *c, char **p, int n)
{ 
   FILE*               file;
   char                str[512];
   struct dupe_s       dp;
   int                 i;
   regex_t             preg;
   int                 type = SEARCH_NORMAL;
   char                *pattern;

   pattern = p[0];
   if (n == 2 && !strcasecmp(p[0], "-e")) {
      if (regcomp(&preg, p[1], REG_EXTENDED|REG_NEWLINE|REG_ICASE) != 0) {
         server_reply(c, "-Invalid regular expression.");
	 return 0;
      }
      type = SEARCH_REGEX;
      pattern = NULL;    
   }
   else if (n == 2 && !strcasecmp(p[0], "-g")) {
      type = SEARCH_GLOB;
      pattern = p[1];
   }
 
   for (i = 0; i < dupe_count; i++) {
      if (is_dupe_detected(type, pattern, (dupe+i)->filename, &preg)) {
         g_snprintf(str, sizeof(str), "+%s", (dupe+i)->filename);
	 server_reply(c, str);
      }
   }
   
   file = open_dupefile("r", 0);
   if (file) {
      while(!feof(file)) {
	 if (fread(&dp, sizeof(dp), 1, file)) {
            if (is_dupe_detected(type, pattern, dp.filename, &preg)) {
               g_snprintf(str, sizeof(str), "+%s", dp.filename);
	       server_reply(c, str);
	    }
	 }
      }
      fclose(file);      
   }
   server_reply(c, "+");
   return 0;
}


/* EOF */
