/* ---------------------------------------------------------------------------
 *
 *   This file is part of OpenFTPD (based on FTP4ALL)
 *   (C) 1996-2000 by Crescent, (C) 2000 by the OpenFTPD Team
 *
 * ---------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 * ---------------------------------------------------------------------------
 *   MODULE: f4adp_ipban.c - all traceroute check / ip ban related stuff
 * ---------------------------------------------------------------------------
 */

#include <sys/types.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include "externsd.h"
#include "log.h"
#include "serverd.h"
#include "utils.h"  
#include "user.h" 
#include "limits.h"
#include "strlcpy.h"
#include "strlcat.h"

struct ip_ident_s
{ char id[32];
  int  version;
  int  count;
  int  temp1;
  int  temp2;
} ip_id = { "OpenFTPD ip file", 0, 0, 0, 0 };

int load_ip_file_first_revision(char *filename, int mode)
{
   FILE                *file;
   struct stat         st;
   char                ipstr[20];
   char                str[20];
   char                *p;
   struct ip_s         *ip = NULL;
   unsigned long       *oldip = NULL;
   int                 count;
   int                 x;
   struct in_addr      host;

   if (-1 == stat(filename, &st) || 0 == st.st_size) {
      return 0;
   }

   if (!(file = fopen(filename, "rb"))) {
      errlog("fopen(%s): %s (%i)", filename, strerror(errno), errno);
      return -1;
   }
   if ((oldip = g_realloc(oldip, st.st_size)) == NULL) {
      fclose(file);
      errlog("load_ip_file_first_revision(): out of memory");
      return -1;
   } 
   count = st.st_size / sizeof(unsigned long);
   if (fread(oldip, sizeof(unsigned long), count, file) != count) {
      fclose(file);
      errlog("load_ip_file_first_revision(): error reading file to memory");
      return -1;
   }
   fclose(file);

   if ((ip = g_realloc(ip, count * sizeof(struct ip_s))) == NULL) {
      errlog("load_ip_file_first_revision(): out of memory");
      return -1;
   }
   memset(ip, 0, count * sizeof(struct ip_s));

   for (x = 0; x < count ; x++) {
      host.s_addr = htonl(oldip[x]);
      g_snprintf(ipstr, sizeof(ipstr), "*@%s", inet_ntoa(host));
      p = ipstr;
      get_iprange(&p, str);
      memcpy((ip + x)->ip, str+12, 8);
   }

   g_free(oldip);

   if (mode) {
      allowip = ip;
      allowip_count = count;
   }
   else {
      denyip = ip;
      denyip_count = count;
   }
   ipbans_dirty++;
   return 0;
}

/* mode: 0 = allow, 1 = deny */
int load_ip_file(int mode)
{ 
   FILE                *file;
   char                str[1024];   
   struct stat         st;
   struct ip_ident_s   ip_test;
   int                 n;
   struct ip_s         *ip = NULL;
   int                 size;
 
   if (mode) mk_filename(str, DFL_HOSTS_ALLOW_FILE, NULL);
   else mk_filename(str, DFL_HOSTS_DENY_FILE, NULL);
   
   /* file does not exist, but this is no fatal error */
   if (-1 == stat(str, &st) || 0 == st.st_size) {
      return 0;
   }
   file = fopen(str, "rb");
   if (!file) {
      errlog("fopen(%s): %s (%i)", str, strerror(errno), errno);
      return -1;
   }
   if (st.st_size < sizeof(ip_test)) {
      errlog("ip file %s hasn't ID, asuming is a first revision one", str);
      fclose(file);
      return load_ip_file_first_revision(str, mode);
   }
   if (!fread(&ip_test, sizeof(ip_test), 1, file)) {
      fclose(file);
      errlog("fread(ip files): %s (%i)",strerror(errno),errno);
      return -1;
   }

   if (strcmp(ip_test.id, ip_id.id)) {
      errlog("ip file %s has bad ID, asuming is a first revision one", str);
      fclose(file);
      return load_ip_file_first_revision(str, mode);
   }
   if (ip_test.version > ip_id.version) {
      errlog("ip file is newer version than daemon");
      fclose(file);
      return -1;
   }
/* For future changes 
   else if (ip_test.version < ip_id.version) {
      errlog("ip file is older version than daemon");
      fclose(file);
      if (ip_test.version == 0) {
         errlog("ip file is 0 version");
         return load_ip_file_v0(str, mode);
      }
      return -1;
   } 
*/ 
   n = ip_test.count;
   if (n < 1) {
      fclose(file);
      return 0;
   }

   if ((ip = g_realloc(ip, n * sizeof(struct ip_s))) == NULL) {
      fclose(file);
      errlog("load_ip_file(): out of memory");
      return -1;
   }
   size = fread(ip, sizeof(struct ip_s), n, file);

   if (size != n) { 
      errlog("load_ip_file(): error reading file to memory");
      fclose(file);
      return -1;
   }

   if (mode) {
      allowip = ip;
      allowip_count = size;
   }
   else {
      denyip = ip;
      denyip_count = size;
   }
   fclose(file);
   ipbans_dirty = 0;
   return 0;
}


int save_ip_file(int mode)
{ 
   FILE*          file;
   char           str[1024];
   char           str2[1030];

   memset(str, 0, sizeof(str));
   memset(str2, 0, sizeof(str2));

   if (mode) mk_filename(str, DFL_HOSTS_ALLOW_FILE, NULL);
   else mk_filename(str, DFL_HOSTS_DENY_FILE, NULL);
   strlcpy(str2, str, sizeof(str2));
   strlcat(str2, ".tmp", sizeof(str2));

   file = fopen(str2, "wb");
   if (file == NULL) { 
      errlog("fopen(%s): %s (%i)", str2, strerror(errno), errno);
      return -1;
   }
   
   if (mode) ip_id.count = allowip_count;
   else ip_id.count = denyip_count;

   if (!fwrite(&ip_id, sizeof(ip_id), 1, file)) {
      errlog("fwrite(%s): %s (%i)", str2, strerror(errno), errno);
      fclose(file);
      return -1;
   }   
   if (fwrite(mode ? allowip : denyip, sizeof(struct ip_s), ip_id.count, file) 
      != ip_id.count) { 
      errlog("fwrite(%s): %s (%i)", str2, strerror(errno), errno);
      fclose(file);
      return -1;
   }
   else { 
      fclose(file);
      if (rename(str2, str) < 0) {
         errlog("rename(%s,%s): %s (%i)", str, str2, strerror(errno), errno);
         return -1;
      }
   }
   return 0;
}


int f4adp_ipban_check(struct child_s *c, char **p, int n)
{
   int             i;
   unsigned long   ipaddr;

   if (!sscanf(p[0], "%lu", &ipaddr)) ipaddr = 0;
   if (!ipaddr) {
      server_reply(c, "-invalid ip");
      return 0;
   }

   /* check the allowed ips first so we can ban a domain 
      and let upload from a single ip of this domain */
   for (i = 0; i < allowip_count; i++) {
      if (in_iprange(ipaddr, (unsigned char*) (allowip+i)->ip)) {
         server_reply(c, "+allowed");
         return 0;
      }
   }

   for (i = 0; i < denyip_count; i++) {
      if (in_iprange(ipaddr, (unsigned char*) (denyip+i)->ip)) {
         server_reply(c, "-denied");
         return 0;
      }
   }

   server_reply(c, "-unknown");
   return 0;
}

/* ban an IP address */
int f4adp_ipban_add(struct child_s *c, char **p, int n)
{
   int             i;
   char            str[20];
   char            ipstr[20];
   char            *po;
   struct ip_s     *ip;

   if (*(p[0] + strlen(p[0]) - 1) == '\n') 
      *(p[0] + strlen(p[0]) - 1) = 0;

   g_snprintf(ipstr,sizeof(ipstr),"*@%s",p[0]);
   po = ipstr;
   if (-1 ==  get_iprange(&po, str)) {
      server_reply(c, "- Invalid IP address");
      return 0;
   }
   print_iprange(ipstr,str);

   /* check in the allowed list, if ip was found we delete it */
   for (i = 0; i < allowip_count; i++) {
      if (!memcmp(str + 12, (allowip+i)->ip, 8)) {
         ip = allowip + i;
         memmove(ip, ip + 1, (--allowip_count-(ip-allowip))*sizeof(struct ip_s));
         ipbans_dirty++;
      }
   }

   /* deny list */
   for (i = 0; i < denyip_count; i++) {
      if (!memcmp(str + 12, (denyip+i)->ip, 8)) {
         server_reply(c, "-This IP address is already banned.");
         return 0;        
      }      
   }
   ip = NULL;
   ip = g_realloc(denyip, (denyip_count + 1) * sizeof(struct ip_s));
   if (!ip) {
      server_reply(c, "-Unable to add IP adress.");
      return 0;        
   }
   denyip = ip;
   memcpy((denyip + denyip_count)->ip, str+12, 8);
   denyip_count++;

   server_log_local(c, LG_SECURITY, "IPBAN of address %s added by %s", 
      ipstr, c->user->name);

   server_reply(c, "+IP address was successfully banned.");
   ipbans_dirty++;
   return 0;         
}


/* unban an IP address */
int f4adp_ipban_rm(struct child_s *c, char **p, int n)
{
   int             i;
   char            str[20];
   char            ipstr[20];
   char            *po;
   struct ip_s     *ip;

   if (*(p[0] + strlen(p[0]) - 1) == '\n') 
      *(p[0] + strlen(p[0]) - 1) = 0;

   g_snprintf(ipstr,sizeof(ipstr),"*@%s",p[0]);
   po = ipstr;
   if (-1 ==  get_iprange(&po, str)) {
      server_reply(c, "- Invalid IP address");
      return 0;
   }
   print_iprange(ipstr,str);

   /* check in the deny list, if ip was found we delete it */
   for (i = 0; i < denyip_count; i++) {
      if (!memcmp(str + 12, (denyip+i)->ip, 8)) {
         ip = denyip + i;
         memmove(ip, ip + 1, (--denyip_count-(ip-denyip))*sizeof(struct ip_s));
         ipbans_dirty++;
      }
   }

   /* allow list */
   for (i = 0; i < allowip_count; i++) {
      if (!memcmp(str + 12, (allowip+i)->ip, 8)) {
         server_reply(c, "-This IP address is already allowed.");
         return 0;        
      }      
   }
   ip = NULL;
   ip = g_realloc(allowip, (allowip_count + 1) * sizeof(struct ip_s));
   if (!ip) {
      server_reply(c, "-Unable to add IP adress.");
      return 0;        
   }
   allowip = ip;
   memcpy((allowip + allowip_count)->ip, str+12, 8);
   allowip_count++;

   server_log_local(c, LG_SECURITY, "IPBAN of address %s removed by %s", 
      ipstr, c->user->name);

   server_reply(c, "+IP address was successfully unbanned.");

   ipbans_dirty++;
   return 0;         
}


int f4adp_ipban_ls(struct child_s *c, char **p, int n)
{
   int             mode;
   char            str[128];
   char            ipstr[16];
   char            ips[20];
   int             i;

   if (!superuser(c->user)) {
      server_reply(c, "-No permission.");
      return 0;   
   }

   if (*p[0] == 'a') mode = IP_ALLOW;
   else if (*p[0] == 'd') mode = IP_DENY;
   else {
      server_reply(c, "-Invalid argument.");
      return 0;
   }

   if (mode == IP_ALLOW) {
      for (i = 0; i < allowip_count; i++) {
         memset(ips, 0, sizeof(ips));
         memcpy(ips + 12, (allowip + i)->ip, 8);
         print_iprange(ipstr, ips);
         g_snprintf(str, sizeof(str), "+%s", ipstr);
 	 server_reply(c, str);
      }   
   }
   else {
      for (i = 0; i < denyip_count; i++) {
         memset(ips, 0, sizeof(ips));
         memcpy(ips + 12, (denyip + i)->ip, 8);
         print_iprange(ipstr, ips);
         g_snprintf(str, sizeof(str), "+%s", ipstr);
 	 server_reply(c, str);
      }      
   }
   server_reply(c, "+");
   return 0;
}



int f4adp_ipban_rst(struct child_s *c, char **p, int n)
{

   if (!superuser(c->user)) {
      server_reply(c, "-No permission.");
      return 0;   
   }

   if (!strncasecmp(p[0], "allow", 5)) {
      g_free(allowip);
      allowip = NULL;
      allowip_count = 0;
   }
   else if (!strncasecmp(p[0], "deny", 4)) {
      g_free(denyip);
      denyip = NULL;
      denyip_count = 0;
   }
   else if (!strncasecmp(p[0], "all", 3)) {
      g_free(allowip);
      allowip = NULL;
      allowip_count = 0;
      g_free(denyip);
      denyip = NULL;
      denyip_count = 0;
   }
   else {
      server_reply(c, "-Invalid argument.");
      return 0;
   }
   server_log_local(c, LG_SECURITY, "IPBAN reset was ordered by %s", 
      c->user->name);
   server_reply(c, "+Reset successful.");
   return 0;
}

/* EOF */
