/*
	$Id: perl_layer.c,v 1.17.2.3 2002/10/26 18:50:01 prime Exp $

	Copyright (C) 2000 Herbert Valerio Riedel <hvr@gnu.org>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "ac_config.h"

#include <stdlib.h>
#include <assert.h>

#if (PATCHLEVEL < 6)
/* needed for ancient perl(tm) */
#undef PACKAGE
#endif

//#include <embed.h>
#include <EXTERN.h>
#include <perl.h>
#include <XSUB.h>
#include <patchlevel.h>

#include <glib.h>

#if (PATCHLEVEL < 6)
#warning enabling workarounds for ancient perl version
#undef eval_pv
#undef get_sv
#undef SvPV_nolen

static SV*
get_sv (const char* name, I32 create)
{
  return perl_get_sv ((char *)name, create);
}

static SV* 
eval_pv (const char *p, I32 croak_on_error)
{
  return perl_eval_pv ((char *)p, croak_on_error);
}

static char *
SvPV_nolen(SV *sv)
{
  STRLEN len;
  return SvPV (sv, len);
}
#endif

#include <glib.h>

#include "perl_layer.h"
#include "control.h"

#include "site_vfs.h"
#include "userdb.h"

extern void expand_variable (char** p, int* len, char* buf, char* str);

static PerlInterpreter *my_perl;  /* The Perl interpreter */

static void export_funcs (void);

extern void boot_DynaLoader (CV *cv);

void
xs_init (void)
{
  gchar *file = __FILE__;
  dXSUB_SYS;
  
  /* DynaLoader is a special case */
  newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
}

void
perl_init (void)
{
  gchar *perl_args[] = {"", "-e", "0"};

  const gchar *sub_load_file =
	"sub FTP::load_file() {"
	" (my $file_name) = @_;"
        " local *FH;"
	" open FH, $file_name or return 2;"
	" local($/) = undef;"
	" my $file = <FH>;"
	" close FH;"
	" eval 'use strict qw(vars); ' . $file;"
        " print (STDERR \"FTP::load_file ('$file_name'):\n\" . $@) if $@;"
	" return 1 if $@;"
	" return 0;"
	"}";

  assert (my_perl == NULL);
  
  my_perl = perl_alloc ();
  perl_construct (my_perl);
  perl_parse (my_perl, xs_init, 3, perl_args, NULL);

  export_funcs ();

  eval_pv(sub_load_file, TRUE);
}

void
perl_done (void)
{
  assert (my_perl != NULL);

  perl_destruct (my_perl);
  perl_free (my_perl);

  my_perl = NULL;
}

static char *
perl_escape (const gchar str[])
{
  GString *tmpstr = NULL;
  gint i = 0;
  
  g_return_val_if_fail (str != NULL, NULL);

  tmpstr = g_string_new (NULL);

  while (str[i])
    switch (str[i])
      {
      case '\0':
	g_assert_not_reached ();
	break;

      case '\\':
      case '\'':
	g_string_append_c (tmpstr, '\\');
      default:
	g_string_append_c (tmpstr, str[i]);
	i++;
	break;
      }

  /* return g_string_free (tmpstr, FALSE); -- GLIB 2.0 */
  {
    gchar *result = tmpstr->str;
    g_string_free (tmpstr, FALSE);
    return result;
  }
}

gint perl_eval_str (const gchar str[])
{
  eval_pv (str, TRUE);
  return 0;
}

static SV *
_perl_call_subfv (const gchar sub[], const gchar fmt[], va_list args)
{
  SV *result;
  GString *perl_cmd = NULL;
  gint i = 0;

  perl_cmd = g_string_new (NULL);
  
  g_string_sprintfa (perl_cmd, "&%s(", sub);

  if (fmt)
    while (fmt[i])
      {
	switch (fmt[i])
	  {
	    gchar *tmp;

	  case 's': /* string */
	    tmp = perl_escape (va_arg (args, const gchar *));
	    g_string_sprintfa (perl_cmd, "'%s'", tmp);
	    g_free (tmp);
	    break;

	  case 'i': /* int */
	    g_string_sprintfa (perl_cmd, "%d", va_arg (args, gint));
	    break;
	  default:
	    g_assert_not_reached ();
	    break;
	  }
	i++;
	if (fmt[i])
	  g_string_append (perl_cmd, ", ");
      }

  g_string_append (perl_cmd, ");");

  result = eval_pv (perl_cmd->str, TRUE);

  g_string_free (perl_cmd, TRUE);

  return result;
}

void perl_set_var_str (const gchar var[], const gchar val[])
{
  SV *_val;

  g_return_if_fail (var != NULL);
  g_return_if_fail (val != NULL);

  _val = get_sv (var, TRUE);

  sv_setpv (_val, val);
}

void perl_set_var_int (const gchar var[], gint val)
{
  SV *_val;
  
  g_return_if_fail (var != NULL);
  
  _val = get_sv (var, TRUE);

  sv_setiv (_val, val);
}

gint
perl_call_subf (const gchar sub[], const gchar fmt[], ...)
{
  va_list args;
  SV *result;

  va_start (args, fmt);
  
  result = _perl_call_subfv (sub, fmt, args);

  va_end (args);
  
  return SvNV (result);
}

const gchar *
perl_call_subf_str (const gchar sub[], const gchar fmt[], ...)
{
  va_list args;
  SV *result;

  va_start (args, fmt);
  
  result = _perl_call_subfv (sub, fmt, args);

  va_end (args);
  
  return SvPV_nolen (result);
}

int 
perl_load_file (const char script_name[])
{
  return perl_call_subf ("FTP::load_file", "s", script_name);
}

/* exported functions */

/* temporary hack */
static const gchar *short_cmds[] =
  { "USR_ACT", "USR_ADQ", "VER", "CNT", "GRP_ADD", 0 };


XS (XS_ftpd_eval)
{
  gchar *eval_str;
  gint cmd_idx = 0;
  gboolean short_cmd = FALSE;
  GString *response;
  dXSARGS;
  items = 0;

  eval_str = SvPV_nolen (ST (0));

  while (short_cmds[cmd_idx])
    if (cmd_idx++, !strncmp (short_cmds[cmd_idx-1], eval_str, 
			     strlen (short_cmds[cmd_idx-1])))
      {
	short_cmd = TRUE;
	break;
      }

  fputs (eval_str, stdout);
  fflush (stdout);

  response = g_string_new (NULL);

  do
    {
      gchar str[128] = { 0, };
    
      if (!fgets (str, sizeof (str), stdin))
	{
	  g_string_append (response, "!Bpipe error!0");
	  break;
	}

      while (*str == '+' && strcmp (str, "+\n")) /* ~= /^+[^\n]/ */
	{
	  g_string_append (response, str + 1);

	  if (short_cmd)
	    break;

	  memset (str, 0, sizeof (str));
	  if (!fgets (str, sizeof (str), stdin))
	    {
	      g_string_append (response, "!Bpipe error, aborting list!0");
	      break;
	    }
	}
      
      if (strcmp (str, "+\n") && !short_cmd)
	g_string_append (response, str + 1);
      
    } while (FALSE);

  XST_mPV (0, response->str);

  g_string_free (response, TRUE);

  XSRETURN (1);
}

XS (XS_get_variable)
{
  gchar *var;
  dXSARGS;
  items = 0;


  var = SvPV_nolen (ST (0));

  {
    gchar buf[1024] = { 0, };
    gchar tmp[3] = "%\0\0";
    gchar *p = tmp + 1;
    *p = var[0];

    expand_variable (&p, NULL, buf, NULL);

    XST_mPV (0, buf);
  }

  XSRETURN (1);
}

XS (XS_userdb_groupname_by_gid)
{
  const gchar *result;
  dXSARGS;
  items = 0;


  if ((result = userdb_groupname_by_gid (SvIV (ST (0)))))
    {
      XST_mPV (0, result);
      XSRETURN (1);
    }
  else
    XSRETURN_EMPTY;
}

XS (XS_userdb_username_by_uid)
{
  const gchar *result;
  dXSARGS;
  items = 0;


  if ((result = userdb_username_by_uid (SvIV (ST (0)))))
    {
      XST_mPV (0, result);
      XSRETURN (1);
    }
  else
    XSRETURN_EMPTY;
}

XS (XS_send_reply)
{
  gchar *msg;
  gint code;
  gint cont;
  dXSARGS;
  items = 0;

  code = SvIV (ST (0));
  msg = SvPV_nolen (ST (1));
  cont = SvIV (ST (2));

  send_reply_full (code, !cont, TRUE, TRUE, msg); /* expand only colors */

  XSRETURN_EMPTY;
}

XS (XS_vfs_getcwd)
{
  dXSARGS;
  items = 0;


  XST_mPV (0, vfs_getcwd ());

  XSRETURN (1);
}

XS (XS_vfs_strerror)
{
  dXSARGS;
  items = 0;


  XST_mPV (0, vfs_strerror (SvIV (ST (0))));

  XSRETURN (1);
}

XS (XS_vfs_readdir)
{
  gchar **files = NULL;
  dXSARGS;
  items = 0;


  if (!(files = vfs_readdir (SvPV_nolen (ST (0)))))
    {
      perl_set_var_int ("FTP::vfs_errno", vfs_errno);
      XSRETURN_EMPTY;
    }
  else
    {
      gint i = 0;

      while (files[i])
	{
	  XST_mPV (i, files[i]);
	  i++;
	}

      g_strfreev (files);

      XSRETURN (i);
    }
}

XS (XS_vfs_stat)
{
  vfs_stat_t st;
  dXSARGS;
  items = 0;

  if (vfs_stat (SvPV_nolen (ST (0)), &st))
    {
      perl_set_var_int ("FTP::vfs_errno", vfs_errno);
      XSRETURN_EMPTY;
    }
  else
    {
      gint i = 0;

      XST_mIV (i, st.type); i++;

      XST_mIV (i, st.uid); i++;
      XST_mIV (i, st.gid); i++;
      XST_mIV (i, st.mode); i++;
      XST_mIV (i, st.cdpath_flag); i++;

      XST_mIV (i, st.create_time); i++;
      XST_mIV (i, st.change_time); i++;
      XST_mIV (i, st.access_time); i++;

      XST_mIV (i, st.size); i++;
      XST_mIV (i, st.crc); i++;
      XST_mIV (i, st.dlc); i++;

      XSRETURN (i);
    }
}

static void
export_funcs (void)
{
  newXS ("FTP::send_reply", XS_send_reply, "FTP");
  newXS ("FTP::ftpd_eval", XS_ftpd_eval, "FTP");
  newXS ("FTP::get_variable", XS_get_variable, "FTP");

  newXS ("FTP::userdb_groupname_by_gid", XS_userdb_groupname_by_gid, "FTP");
  newXS ("FTP::userdb_username_by_uid", XS_userdb_username_by_uid, "FTP");

  newXS ("FTP::vfs_strerror", XS_vfs_strerror, "FTP");
  newXS ("FTP::vfs_stat", XS_vfs_stat, "FTP");
  newXS ("FTP::vfs_getcwd", XS_vfs_getcwd, "FTP");
  newXS ("FTP::vfs_readdir", XS_vfs_readdir, "FTP");
}

/* 
 * Local variables:
 *	c-file-style: "gnu"
 *	tab-width: 4
 *	indent-tabs-mode: t
 * End:
 */
