/* ---------------------------------------------------------------------------
 *
 *   This file is part of OpenFTPD (based on FTP4ALL)
 *   (C) 1996-2000 by Crescent, (C) 2000 by the OpenFTPD Team
 *
 * ---------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 * ---------------------------------------------------------------------------
 *   MODULE: perm.c - permission related stuff
 * ---------------------------------------------------------------------------
 */

#include "ac_config.h"

#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <utime.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include "externs.h"
#include "perm.h"
#include "utils.h"
#include "site_vfs.h"

int superuser(void)
{ return (usr.class & UC_MASK) == UC_ROOT;
}

int groupop(void)
{ return (usr.class & UC_MASK) == UC_GROUPOP;
}

int siteop(void)
{ return (usr.class & UC_MASK) == UC_SITEOP;
}

int invisible(void)
{ return (usr.class & UF_INVISIBLE);
}

int coloruser(void)
{ return (usr.class & UF_COLOR);
}


int in_group(gint32 gid)
{ int i;

  if (usr.gid == gid) return 1;
  for (i=0; i<8; i++) if ((usr.gids[i] != -1) && (usr.gids[i] == gid))
    return 1;
  return 0;
}

gint32 get_gid(char* name)
{ struct group_s* g = group;
  gint32           gid;
  int             i;

  for (i=0; i<group_count; i++)
  { if (!strcasecmp(g->name,name)) return g->gid;
    g++;
  }
  if (sscanf(name,SI32,&gid)) return gid; else return -1;
}

gint32 get_uid(char* name)
{ int            i;
  struct user_s* u = user;
  gint32          uid;

  for (i=0; i<user_count; i++)
  { if (!strcasecmp(u->name,name)) return u->uid;
    u++;
  }
  if (sscanf(name,SI32,&uid)) return uid; else return -1;
}

struct group_s* get_group_by_gid(gint32 gid)
{ struct group_s* g = group;
  int             i;

  for (i=0; i<group_count; i++, g++) if (g->gid == gid) return g;
  return NULL;
}

char* get_groupname(gint32 gid)
{ struct group_s* g;
  static char*    incoming = "incoming";
  static char*    nogroup = "nogroup";

  if (gid == -1) return nogroup;
  if (gid == -2) return incoming;
  g = get_group_by_gid(gid);
  if (g) return g->name; else return NULL;
}

struct user_s* get_user_by_uid(gint32 uid)
{ int            i = 0;

  for (i = 0; i < user_count; i++)
    if (((user + i)->uid) == uid)
      return(user + i);
  return NULL;
}

char* get_username(gint32 uid)
{ static char*   incoming = "incoming";
  static char*   nobody = "nobody";
  struct user_s* u;

  if (uid == -1) return nobody;
  if (uid == -2) return incoming;
  u = get_user_by_uid(uid);
  if (u) return u->name; else return NULL;
}

void split_dir(char* fp, char** dir, char** name, int testdir)
{ 
   char*       p;
   struct stat s;
   char        wd[256];

   if (testdir) {
      if (!add_dir(vfs_getcwd (),fp,wd,sizeof(wd)) &&
         !stat(strcmp(wd, "/") ? wd+1 : ".", &s)) { 
      
	 if (s.st_mode & S_IFDIR) { 
            *dir = fp;
            *name = fp+strlen(fp);
            return;
	 }
      }
   }
   if (*fp == '/') { 
      p = strrchr(fp+1,'/');
      if (!p) { 
         *dir = NULL;
         *name = fp+1;
      } 
      else { 
         *p = '\0';
         *dir = fp;
         *name = p+1;
      }
   } 
   else { 
      p = strrchr(fp,'/');
      if (!p) { 
         *dir = fp+strlen(fp);
         *name = fp;
      } 
      else { 
         *p = '\0';
         *dir = fp;
         *name =  p+1;
      }
   }
}

FILE* file_open(const char* dir, char* filename, const char* mode, gint32 pcheck)
{ 
   struct stat    st;
   struct utimbuf t;
   char           wd[256];
   char*          hfile;
   char           str[256];

   if (add_dir(vfs_getcwd (),dir,wd,sizeof(wd)) == -1) return NULL;
   if (strlen(wd)+strlen(filename)+1 >= sizeof(wd)) { 
      errno = ENAMETOOLONG;
      return NULL;
   }
   if (strcmp(wd,"/")) strncat(wd, "/", sizeof(wd) - 1);
   strncat(wd, filename, sizeof(wd) - 1);
   if (pcheck && !superuser()) { 
      if (!strcmp(filename,permission_file)) { 
         errno = EACCES;
         return NULL;
      }
      if (readme_file && !strcmp(filename,readme_file)) { 
         errno = EACCES;
         return NULL;
      }

      /* added by primemover */
      strncpy(str, hidden_files, sizeof(str) - 1);
      hfile = strtok(str, " \t");
      while (hfile != NULL) {
         if (!strcmp(filename, hfile)) {
	    errno = EACCES;
	    return NULL;	    
         }
         hfile = strtok(NULL, " \t");
      }

    /***********************/
    if (pcheck == PE_READ)
    { if (stat(wd+1,&st) == -1) return NULL;
      if (permission_ok(dir,PE_CHDIR) == -1) return NULL;
      if (file_permission_ok(dir,filename,FPE_READ) == -1) return NULL;
    } else
    { if (stat(wd+1,&st) == -1)
      { if (permission_ok(dir,PE_WRITE) == -1) return NULL;
      } else
      { if (permission_ok(dir,PE_CHDIR) == -1) return NULL;
        if (file_permission_ok(dir,filename,FPE_WRITE) == -1) return NULL;
      }
    }
  }
  if (flag_get(flags,FLAG_TOUCH))
  { t.modtime = time(&t.actime);
    utime(wd+1,&t);
  }
  return fopen(wd+1,mode);
}



/* returns:
       -1 ... permission denied
     >= 0 ... ok (returns directory flags)
 */
int permission_part_ok(const char* dir, gint32 p, int uidok)
{ 
   gint32 gid;
   gint32 perm;
   gint32 uid;

   if (read_permission_file(dir, &uid, &gid, &perm, NULL, NULL, NULL) == -1) {
      if (superuser()) return 0;
      errno = EACCES;
      return -1;
   }
   if (superuser()) 
      return (perm >> 24) & 0xFF;
   /* if (uidok) if (usr.uid == uid) return (perm >> 24) & 0xFF;*/
   errno = EACCES;
   if (p & PE_CHMOD) 
      if (usr.uid != uid) 
         return -1;
   if (usr.uid == uid) 
      if (((perm >> 16) & p) == p) 
         return (perm >> 24) & 0xFF;
   if (in_group(gid)) 
      if (((perm >> 8) & p) == p) 
         return (perm >> 24) & 0xFF;
   if ((perm & p) == p) 
      return (perm >> 24) & 0xFF;
   return -1;
}

int directory_hidden(char* dir)
{ gint32 gid;
  gint32 perm;
  gint32 uid;

  if (superuser()) return 0;
  if (read_permission_file(dir,&uid,&gid,&perm,NULL,NULL,NULL) < 0) return 0;
  if ((perm >> 24) & DP_HIDDEN)
  { if (usr.uid == uid) if ((perm >> 16) & PE_CHDIR) return 0;
    if (in_group(gid)) if ((perm >> 8) & PE_CHDIR) return 0;
    if (perm & PE_CHDIR) return 0;
    return 1;
  }
  return 0;
}
#if 1
gboolean
directory_flag_set_recursive (const char* dir, int flag)
{
  vfs_stat_t st;

  g_return_val_if_fail (flag == VFS_DF_HIDDEN, FALSE);

  if (vfs_stat (dir, &st))
    {
      g_printerr ("directory_flag_set_recursive(%s): %s\n", dir, vfs_strerror (vfs_errno));
      return FALSE;
    }

  return (vfs_perm_get (st.mode, VFS_FLAG) & VFS_DF_HIDDEN
	  || vfs_perm_get (st.mode, VFS_FLAG) & VFS_DF_HIDDEN_R);
}
#else
int
directory_flag_set_recursive(const char* dir, int flag)
{
  gint32 gid;
  gint32 perm;
  gint32 uid;
  char cwdorg[256];
  char fullpath[256];
  char *p;

  g_printerr ("fixme directory_flag_set_recursive()\n");

  add_dir(vfs_getcwd (), dir, fullpath, sizeof(fullpath));
  strncpy(cwdorg, vfs_getcwd (), sizeof(cwdorg) - 1);
  vfs_setcwd ("/");
  while (strcmp(fullpath, "/") && (p = strrchr(fullpath, '/')) != NULL ) 
    {
      if (read_permission_file(fullpath,&uid,&gid,&perm,NULL,NULL,NULL) == 0)
	if ((perm >> 24) & flag) 
	  {
	    vfs_setcwd (cwdorg);
	    return 1;
	  }
      *p = 0;
    }
  vfs_setcwd (cwdorg);
  return 0;
}
#endif

int directory_flag_set(char* dir, int flag)
{ gint32 gid;
  gint32 perm;
  gint32 uid;

  if (read_permission_file(dir,&uid,&gid,&perm,NULL,NULL,NULL) < 0) return 0;
  if ((perm >> 24) & flag) return 1;
  return 0;
}


int permission_ok(const char* dir, gint32 perm)
{ 
   char        c;
   int         n;
   char*       p;
   struct stat st;
   char        wd[256];

   if ((perm < 0) || (perm > 255)) { 
      errno = EACCES;
      return -1;
   }
   if (!dir) strncpy(wd, "/", sizeof(wd) - 1); 
   else if (add_dir(vfs_getcwd (), dir, wd, sizeof(wd)) == -1) return -1;
   if (strcmp(wd, "/")) { 
      if (stat(wd+1, &st) == -1) return -1;
      if (!S_ISDIR(st.st_mode)) { 
         errno = ENOTDIR;
         return -1;
      }
   }
   if (superuser()) return permission_part_ok(wd, perm, 0);
   p = wd+1;
   /* check all parent dirs */
   do { 
      while (*p && (*p != '/')) p++;
      c = *p;
      *p = '\0';
      if (c && *(p+1)) { 
         n = permission_part_ok(wd, PE_CHDIR, 0);
	 *(p++) = c;
	 if (n == -1) return -1;
      } 
      else { 
         n = permission_part_ok(wd, perm, 0);
	 *p = c;
	 if (n == -1) return -1;
	 if (c) p++;
      }
   } while (*p);
   return n;
}






int file_permission_ok(const char* dir, char* name, gint32 tperm)
{ 
   gint32 gid;
   gint32 perm;
   gint32 uid;
   char* hfile;
   char  str[256];

   /* added by primemover */
   if (*name && (!strcmp(permission_file, name) || !strcmp(readme_file, name))) {
      errno = EACCES;
      return -1;
   }

   strncpy(str, hidden_files, sizeof(str) - 1);
   hfile = strtok(str, " \t");
   while (hfile != NULL) {
      if (*name && (!strcmp(hfile, name))) {
         errno = EACCES;
  	 return -1;
      }
      hfile = strtok(NULL, " \t");
  }
 
  /* siteops are allowed to delete too now! */
  /* if (superuser()) return 0; */
  if (superuser() || siteop()) return 0;
  if (read_file_permission(dir,name,&uid,&gid,&perm) == -1) return -1;
  if (uid == usr.uid) if ((((perm >> 16) & FP_ALL) & tperm) == tperm)
    return 0;
  if (in_group(gid)) if ((((perm >> 8) & FP_ALL) & tperm) == tperm)
    return 0;
  if (((perm & FP_ALL) & tperm) == tperm) return 0;
  errno = EACCES;
  return -1;
}

/* EOF */
