#!/usr/bin/perl -w

##############################################################################
#                                                                            #
#  primeSTATS v2.2.1 - (c) 2000 primemover                                   #
#                                                                            #
#  frontend for OpenFTPD statfiles                                           #
# ---------------------------------------------------------------------------#
#                                                                            #
#  This program is free software; you can redistribute it and/or modify      #
#  it under the terms of the GNU General Public License as published by      #
#  the Free Software Foundation; either version 2 of the License, or         #
#  (at your option) any later version.                                       #
#                                                                            #
##############################################################################

my $debug = 0;

my $userstats_body  = " !C|!H %02d %-9.9s %-9.9s %-24.24s %8.0f MB %5.0f KB/s !C|!0\n";
my $groupstats_body = " !C|!H %02d %-9.9s %-34.34s %8.0f MB %5.0f KB/s !C|!0\n";


##############################################################################

use strict; #always
use FindBin;
use lib "$FindBin::Bin/";
use primetools;
use Cwd;
use File::Basename;

# Shared vars
our (%stats,$user,$statdir,$group,$section,$lines,$filename,$pid);

sub load_stats {
   my $statfile = "$standarddir/$statdir/" . lc($section) . ".txt";
   if ($debug) { print "DEBUG: $statfile\n"; }
   open(STATFILE, "< $statfile") or return 0;
   if ($debug) { print "DEBUG: statfile opened sucessfully\n"; }
   while (<STATFILE>) {
      if ($debug) { print "DEBUG: 1 line read from statfile\n"; }
      my @s = split(/\t/, $_);
      if ($#s != 33) {
         die "unknown file format.";
      }
      my $username = $s[0];
      $stats{$username}{userinfo}    = $s[1];
      $stats{$username}{groupname}   = $s[2];
      $stats{$username}{groupinfo}   = $s[3];
      $stats{$username}{dl_f_day}    = $s[4];
      $stats{$username}{dl_day}      = $s[5];
      $stats{$username}{dl_tm_day}   = $s[6];
      $stats{$username}{ul_f_day}    = $s[7];
      $stats{$username}{ul_day}      = $s[8];
      $stats{$username}{ul_tm_day}   = $s[9];
      $stats{$username}{dl_f_week}   = $s[10];
      $stats{$username}{dl_week}     = $s[11];
      $stats{$username}{dl_tm_week}  = $s[12];
      $stats{$username}{ul_f_week}   = $s[13];
      $stats{$username}{ul_week}     = $s[14];
      $stats{$username}{ul_tm_week}  = $s[15];
      $stats{$username}{dl_f_month}  = $s[16];
      $stats{$username}{dl_month}    = $s[17];
      $stats{$username}{dl_tm_month} = $s[18];
      $stats{$username}{ul_f_month}  = $s[19];
      $stats{$username}{ul_month}    = $s[20];
      $stats{$username}{ul_tm_month} = $s[21];
      $stats{$username}{dl_f_year}   = $s[22];
      $stats{$username}{dl_year}     = $s[23];
      $stats{$username}{dl_tm_year}  = $s[24];
      $stats{$username}{ul_f_year}   = $s[25];
      $stats{$username}{ul_year}     = $s[26];
      $stats{$username}{ul_tm_year}  = $s[27];
      $stats{$username}{dl_f_total}  = $s[28];
      $stats{$username}{dl_total}    = $s[29];
      $stats{$username}{dl_tm_total} = $s[30];
      $stats{$username}{ul_f_total}  = $s[31];
      $stats{$username}{ul_total}    = $s[32];
      $stats{$username}{ul_tm_total} = $s[33];
   }
   close(STATFILE);
   return 1;
}

sub print_user_stats {
   my $v1 = $_[0] . '_' . $_[1];
   my $v2 = $_[0] . '_tm_' . $_[1];
   my $v3 = $_[0] . '_f_' . $_[1];
   my $rank = 0;
   foreach $user (sort {$stats{$b}{$v1} <=> $stats{$a}{$v1}} keys %stats) {   
      my $mbsize = $stats{$user}{$v1} / 0x100000;
      my $speed = $stats{$user}{$v2} > 0 ? 
                  $stats{$user}{$v1} / ($stats{$user}{$v2} * 1024) : 0;
      my $files = $stats{$user}{$v3};
      my $group = $stats{$user}{groupname};
      my $info  = $stats{$user}{userinfo};
      $rank++;
      if ($rank > $lines) { last; }
      printf $userstats_body, $rank, $user, $group, $info, $mbsize, $speed;
   }
}

# ----------------------------------------------------------------------------
#  print_group_stats(<ul|dl>, <day|week|month|year|total>)
# ----------------------------------------------------------------------------
sub print_group_stats {
   my %groupstats = ();
   my $v1 = $_[0] . '_' . $_[1];
   my $v2 = $_[0] . '_tm_' . $_[1];
   my $rank = 0;
   foreach $user (keys %stats) {
      my $groupname = $stats{$user}{groupname};
      if (exists $groupstats{$groupname}) {
         $groupstats{$groupname}{$v1} += ($stats{$user}{$v1});
         $groupstats{$groupname}{$v2} += ($stats{$user}{$v2});
      }
      else {
	 $groupstats{$groupname}{groupinfo} = $stats{$user}{groupinfo};
         $groupstats{$groupname}{$v1} = $stats{$user}{$v1};
         $groupstats{$groupname}{$v2} = $stats{$user}{$v2};
      }
   }
   foreach $group (sort {$groupstats{$b}{$v1} <=> $groupstats{$a}{$v1}} keys %groupstats) {
      my $mbsize = $groupstats{$group}{$v1} / 0x100000;
      my $speed = $groupstats{$group}{$v2} > 0 ? 
                  $groupstats{$group}{$v1} / ($groupstats{$group}{$v2} * 1024) : 0;
      my $info = $groupstats{$group}{groupinfo};
      $rank++;
      if ($rank > $lines) { last; }
      printf $groupstats_body, $rank, $group, $info, $mbsize, $speed;
   }
}


sub print_header {
   open(HEADERFILE, "< ${standarddir}/msg/stat_top") 
   or die "Unable to open headerfile $filename";
   while(<HEADERFILE>) {
      if ($_[0] =~ /^u../) { 
         $_ =~ s/\$USRGRP/USER   /;
      }
      else {
         $_ =~ s/\$USRGRP/GROUP  /;
      }

      if ($_[0] =~ /^.d./) { 
         $_ =~ s/\$TIME/DAY  /;
      }
      elsif ($_[0] =~ /^.w./) { 
         $_ =~ s/\$TIME/WEEK /;
      }
      elsif ($_[0] =~ /^.m./) { 
         $_ =~ s/\$TIME/MONTH/;
      }
      elsif ($_[0] =~ /^.y./) { 
         $_ =~ s/\$TIME/YEAR /;
      }
      else {
         $_ =~ s/\$TIME/TOTAL/;
      }

      if ($_[0] =~ /^..u/) { 
         $_ =~ s/\$DIRECTION/UPLOAD    /;
      }
      else {
         $_ =~ s/\$DIRECTION/DOWNLOAD  /;
      }
      print $_;
   }
   close(HEADERFILE);
   $section = sprintf("%-17.17s", uc($section));
   print " !C| !hSECTION: !H$section                                             !C|!0\n";
   print " !C|------------------------------------------------------------------------|!0\n";
   if ($_[0] =~ /^u../) { 
      print " !C|!3!H ## User      Group     Infoline                   Megabytes  Avg-Speed !0!C|!0\n";
   }
   else {
      print " !C|!3!H ## Group     Infoline                             Megabytes  Avg-Speed !0!C|!0\n";
   }
   print " !C|------------------------------------------------------------------------|!0\n";
}


sub print_footer {
   open(FILE, "< ${standarddir}/msg/stat_tail") 
   or die "Unable to open $filename";
   while(<FILE>) {
      print $_;
   }
   close(FILE);
}

sub refresh_stats {
   open(PIDFILE, "< ${standarddir}/var/ftpd.pid");
   $pid = <PIDFILE>;
   chomp($pid);
   kill USR1 => $pid;
   sleep(1);
}

###############################################################################
# main program                                                                #
###############################################################################

$statdir = "var/stats";
if (exists $ENV{'F4A_STD'}) {
  $standarddir = $ENV{'F4A_STD'};
}
elsif ( -e "$ENV{'HOME'}/.openftpdrc" ) { 
  $standarddir = `cat $ENV{'HOME'}/.openftpdrc`;
  $standarddir =~ s/\n//;
  $standarddir =~ s/\r//;
}
else {
  print "ERROR: Cant locate your standarddir either in Environment F4A_STD or in $ENV{'HOME'}/.openftpdrc\n"; 
  exit 1;
}

%stats = ();
$lines = 10;

if (exists $ENV{'F4A_SECTION'}) { 
   $section = lc($ENV{'F4A_SECTION'});
}
else { 
   # couldnt read the section, so lets use global, instead of exiting...
   $section = "global";
#   print "ERROR: Unable to read F4A_SECTION\n";
#   exit 1;
};

# we want to see the global stats and not the default one
if ($section eq "default") {
   $section = "global";
}

if ($#ARGV < 0 || $#ARGV > 2) {
   print "OpenFTPD stat script (c) 2000 primemover\n";
   print "Syntax: stats.pl <u|g><d|w|m|y|t><u|d> [section] [lines]\n";
   print "        Example: stats.pl uwu iso 10\n";
   exit;
}

if ($#ARGV > 0) {
   $section = $ARGV[1];
   chomp($ARGV[1]);
}

if ($#ARGV > 1 && $ARGV[2] =~ /\d+/) {
   $lines = $ARGV[2];
}
refresh_stats();
load_stats();

if ($ARGV[0] =~ /(u|g)(d|w|m|y|t)(u|d)/) {
  if ($1 eq "u") {
     if ($2 eq "d") {
        if ($3 eq "u") {
           print_header("udu");
           print_user_stats("ul", "day");
           print_footer("udu");	
	
	}
	elsif ($3 eq "d") {
           print_header("udd");
           print_user_stats("dl", "day");
           print_footer("udd");	
	}
     }
     elsif ($2 eq "w") {
        if ($3 eq "u") {
           print_header("uwu");
           print_user_stats("ul", "week");
           print_footer("uwu");	
	
	}
	elsif ($3 eq "d") {
           print_header("uwd");
           print_user_stats("dl", "week");
           print_footer("uwd");	
	}
     }
     elsif ($2 eq "m") {
        if ($3 eq "u") {
           print_header("umu");
           print_user_stats("ul", "month");
           print_footer("umu");		
	}
	elsif ($3 eq "d") {
           print_header("umd");
           print_user_stats("dl", "month");
           print_footer("umd");		
	}    
     }
     elsif ($2 eq "y") {
        if ($3 eq "u") {
           print_header("uyu");
           print_user_stats("ul", "year");
           print_footer("uyu");		
	}
	elsif ($3 eq "d") {
           print_header("uyd");
           print_user_stats("dl", "year");
           print_footer("uyd");		
	}     
     }
     elsif ($2 eq "t") {
        if ($3 eq "u") {
           print_header("utu");
           print_user_stats("ul", "total");
           print_footer("utu");		
	
	}
	elsif ($3 eq "d") {
           print_header("utd");
           print_user_stats("dl", "total");
           print_footer("utd");			
	}    
     }
  }
  elsif ($1 eq "g") {
     if ($2 eq "d") {
        if ($3 eq "u") {
           print_header("gdu");
           print_group_stats("ul", "day");
           print_footer("gdu");	
	
	}
	elsif ($3 eq "d") {
           print_header("gdd");
           print_group_stats("dl", "day");
           print_footer("gdd");	
	}
     }
     elsif ($2 eq "w") {
        if ($3 eq "u") {
           print_header("gwu");
           print_group_stats("ul", "week");
           print_footer("gwu");	
	
	}
	elsif ($3 eq "d") {
           print_header("gwd");
           print_group_stats("dl", "week");
           print_footer("gwd");	
	}
     }
     elsif ($2 eq "m") {
        if ($3 eq "u") {
           print_header("gmu");
           print_group_stats("ul", "month");
           print_footer("gmu");		
	}
	elsif ($3 eq "d") {
           print_header("gmd");
           print_group_stats("dl", "month");
           print_footer("gmd");		
	}    
     }
     elsif ($2 eq "y") {
        if ($3 eq "u") {
           print_header("gyu");
           print_group_stats("ul", "year");
           print_footer("gyu");		
	}
	elsif ($3 eq "d") {
           print_header("gyd");
           print_group_stats("dl", "year");
           print_footer("gyd");		
	}     
     }
     elsif ($2 eq "t") {
        if ($3 eq "u") {
           print_header("gtu");
           print_group_stats("ul", "total");
           print_footer("gtu");		
	
	}
	elsif ($3 eq "d") {
           print_header("gtd");
           print_group_stats("dl", "total");
           print_footer("gtd");			
	}    
     }
   }  
}

# --- end of file ---
